/*
 
 DESCRIPTION:
 Include file for GPIBCON.
 
 Copyright 2019 Brian Cornell.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 of the Software, and to permit persons to whom the Software is furnished to do
 so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 DEPENDENCIES:  none
 
 VERSION:  2.0
 
 MODIFICATION HISTORY:
 
 24 Mar 19, BWC, second release.
 13 Feb 19, BWC, first release.
 
 
 */


#define VERMAJOR        2
#define VERMINOR        0
#define VERBUILD        7

#define TRUE            1
#define FALSE           0

#define SYS_EPOCH_YEAR  1900            //Epoch for C library
#define LOG_EPOCH_YEAR  2000            //Epoch for logger HW

#define FMODE_GPIB      "r+"            //Need r/w
#define FMODE_CONF      "r"

#define BASE_TEN        10

#define POS_ARG_PRG     0
#define DSP_POS_VAL_R   3
#define DSP_POS_VAL_C   1
#define DSP_POS_CMD_R   15              //Command prompt
#define DSP_POS_CMD_C   0               //Command prompt
#define DSP_POS_RSP_R   16              //Command response
#define DSP_POS_RSP_C   0               //Command response
#define DSP_POS_EXT_R   24              //Quit requested from console
#define DSP_POS_EXT_C   79              //Quit requested from console

//Execution delay constants
#define DLY_S_INT_C     0               //# S between initialization commands
#define DLY_NS_INT_C    20000000        //# nS between initialization commands
#define DLY_S_RD_C      0               //# S between read commands
#define DLY_NS_RD_C     1000000         //# nS between read commands
#define DLY_S_DUP_C     0               //# S between display updates
#define DLY_NS_DUP_C    900000000       //# nS between display updates
#define DLY_S_OFT_C     1               //# S between off commands
#define DLY_NS_OFT_C    0               //# nS between off commands
#define DLY_S_OFF_C     0               //# S between off commands
#define DLY_NS_OFF_C    20000000        //# nS between off commands

//Ansi terminal control & characters
#define ANSIESC         "\e["
#define ANSICLRSCR      "2J"
#define ANSICPOS        'H'
#define ANSIELINE       'K'
#define ANSISAVE        "\e7"
#define ANSIRESTORE     "\e8"
#define ACP_SZ_STR      4
#define ACP_MAX_ROW     25
#define ACP_MAX_COL     80
#define CHR_CR          0x0D
#define CHR_LF          0x0A
#define CHR_NULL        0x00
#define CHR_COLON       ':'
#define CHR_SEMICOLON   ';'
#define CHR_SPACE       ' '

//Dimensions
#define SIZ_BUF_CFG     32
#define SIZ_BUF_CMD     32
#define SIZ_BUF_HLP     64
#define SIZ_BUF_RSP     64              //GPIB response buffer
#define SIZ_BUF_VAR     64              //Should be >= largest buffer in gcmd
#define SIZ_ARY_IDX     10
#define SIZ_ARY_DCT     50
#define SIZ_BUF_FILE    256
#define SIZ_ARGC_MIN    2               //Minimum # arguments
#define SIZ_BUF_CON     48              //User console (command)
#define SIZ_CMD_FLAG    2               //Size of command line flag
#define SIZ_ARY_FLD     2               //Array holding float values of math fields

//Array addresses
#define ARY_FLD_FLD1    0               //First math field
#define ARY_FLD_FLD2    1               //Second math field

//Program strings
#define STR_PRGNAME     "gpibcon"       //Program name
#define STR_EXT_CFG     ".conf"         //Config extension
#define STR_ERROR       "ERR"           //Error

//Print formats
#define FMT_INT_WR      "%s\n"          //End initialization command with NL
#define FMT_OFF_WR      "%s\n"          //End off command with NL
#define FMT_DSP_WR      "%s\n"          //End display command with NL
#define FMT_DSP_VAL     "%s%s\t"        //Display value format
#define FMT_DMP_HDR     "\n---%s section---\n"
#define FMT_DMP_INT     "cmd=%s\n"
#define FMT_DMP_RD      "cmd=%s\n"
#define FMT_DMP_CFG     "cfg=%s, cmd=%s, usrv=%d, crsp=%d, hlp=%s\n"
#define FMT_DMP_DSP     "%d: label=%s, cmd=%s, row=%d, rltx=%d, info=%s\n"
#define FMT_DMP_MTH     "\top=%c, f1=%d, f2=%d\n"
#define FMT_DMP_OFF     "cmd=%s\n"
#define FMT_DMP_END     "\n# records=%d, max=%d\n"
#define FMT_CMD_PR      ": "
#define FMT_RSP_PR      "=>"
#define FMT_RSP_NOV     "Value required"
#define FMT_EXC_CMD     "%s %s\n"
#define FMT_RSP_IOE     "device io error"
#define FMT_EXC_RSP     "%s"
#define FMT_RSP_INV     "Invalid command"
#define FMT_HLP_DCM     "\n++++++++++Dictonary Commands++++++++++\n"
#define FMT_HLP_ICM     "\n++++++++++Internal Commands++++++++++\n"
#define FMT_HLP_TXT     "%s: %s\n"
#define FMT_HLP_END     "\n\nPress ENTER to resume."
#define FMT_ERR_STDERR  "\n%s:  %s, errno=%d."
#define FMT_TRM_SIGNAL  "\n%s terminating on signal %d.\n"
#define FMT_PRG_DESC    "\nGPIB console interface, v%d.%d build %d.  \
(c) 2019 Brian Cornell.\n"
#define FMT_PRG_SYNTX   "Syntax:  %s (flags) /dev/<device>.  Optional flags:"
#define FMT_PRG_CMDL    "\n\t%s - %s"
#define FMT_PRG_EXIT    "\n\n"
#define FMT_MTH_RESULT  "%06.6f"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Constants for gcmd objects
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Type gcmd object types
#define REC_TYP_READ    0               //Command to read value
#define REC_TYP_INIT    1               //Initialization command
#define REC_TYP_CFG     2               //User entered configuration commands
#define REC_TYP_DSP     3               //Display commands
#define REC_TYP_OFF     4               //Shutdown commands

//Field positions for cfg records
#define POS_CFG_USRC    1               //User (console) command name
#define POS_CFG_CMDC    2               //GPIB (machine) command name
#define POS_CFG_PARM    3               //Command parameter required
#define POS_CFG_CRSP    4               //Read command response
#define POS_CFG_HELP    5               //Help string

//Field positions for dsp records
#define POS_DSP_LABL    1               //Display label
#define POS_DSP_CMDC    2               //GPIB (machine) command name
#define POS_DSP_ROW     3               //Display row
#define POS_DSP_RLTX    4               //Remove leading text from response
#define POS_DSP_MATH    5               //Math operator
#define POS_DSP_FLD1    6               //1st operand field
#define POS_DSP_FLD2    7               //2nd operand field
#define POS_DSP_HELP    8               //Help string

//Config file command prefixes
#define REC_NAM_INIT    "init:"         //Initialization
#define REC_NAM_READ    "read:"         //Read
#define REC_NAM_CFG     "cfg:"          //Configuration
#define REC_NAM_DSP     "dsp:"          //Display
#define REC_NAM_OFF     "off:"          //Shutdown

//Display record command tags
#define DSP_TAG_CALC    "$calc$"        //If found in command field indicates calculated field

//Dictionary index pointers
#define DICT_INIT_F     0
#define DICT_INIT_L     1
#define DICT_READ_F     2
#define DICT_READ_L     3
#define DICT_CONF_F     4
#define DICT_CONF_L     5
#define DICT_DISP_F     6
#define DICT_DISP_L     7
#define DICT_OFF_F      8
#define DICT_OFF_L      9

//Math operators
#define MATH_OP_BAD     0               //Invalid
#define MATH_OP_ADD     1               //Add
#define MATH_OP_SUB     2               //Subtract
#define MATH_OP_MUL     3               //Multiply
#define MATH_OP_DIV     4               //Divide

//Math operator symbols array
#define MATH_ARY_OPS    " +-*/"

//Dictionary object
typedef struct {
    int   type;                         //Object type
    int   usrv;                         //type=cfg value req; type=dsp then row
    int   crsp;                         //type=cfg get command response, type=dsp remove leading text
    int   math;                         //type=dsp see MATH_OP_XXX constants
    int   fld[SIZ_ARY_FLD];             //Display fields to perform calc on, offset-1
    void *prev;                         //Pointer to previous object of type
    void *next;                         //Pointer to next object of type
    char  cfg[SIZ_BUF_CFG];             //type=cfg, user command, type=dsp then label
    char  cmd[SIZ_BUF_CMD];             //Machine version of command
    char  hlp[SIZ_BUF_HLP];             //type=cfg Help, type=dsp label info
    char  rsp[SIZ_BUF_RSP];             //GPIB response
} gcmd;


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Constants for cmd objects
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Command object
typedef struct {
    const char *str;
    const char *desc;
    char       *value;
} cmd;

//Internal command strings
#define CMD_A_QUIT      0
#define CMD_A_CLS       1
#define CMD_A_HELP      2

const cmd Icmd[] = {
    {"quit","Quit program"},
    {"cls","Redraw screen"},
    {"help","Display command help"}
};

#define SIZ_A_ICMD      sizeof(Icmd) / (sizeof(const char *) * 3)


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Command line processing
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Flag bit mapping
#define FLG_NONE        0x0000
#define FLG_BIT_SET     0x0001
#define FLG_DIFF_CONF   0x0001
#define FLG_DUMP_CONF   0x0002

//Flag array position
#define FLG_A_DIFF_CONF 0
#define FLG_A_DUMP_CONF 1

cmd Flags[] = {
    {"-c","use different config file"},
    {"-d","dump config file & exit"}
};

#define SIZ_A_FLG       sizeof(Flags) / (sizeof(const char *) * 3)


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Return codes & messages
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define RC_SUCCESS      0
#define RC_NOARG        1
#define RC_BADARG       2
#define RC_BADCFG       3
#define RC_DEVOFAIL     4
#define RC_SIGERROR     5
#define RC_INITFAIL     6
#define RC_THREAD       7
#define RC_SIGEXIT      8

const char *retCode[] = {
    "success",
    "no arguments",
    "bad or missing arguments",
    "problem with config file",
    "can't open device",
    "error installing signal handler",
    "error initializing device",
    "problem starting display thread",
    "CTRL-C or TERM signal received"
};


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Function declarations
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int main(int, char *[]);
int cmdProc(void);
int loadCfg(char *, int, gcmd *[], gcmd *);
void dumpCfg(gcmd *[], gcmd *);
int gpibInit(FILE *, gcmd *);
int gpibOff(FILE *, gcmd *);
void gpibDsp(void *);
int strnl(unsigned char *, int);
int strlnshift(unsigned char *, int);
int strlrtrim(unsigned char *, int);
void sigHandler(int);
const char *getCfgVar(char *, int);
void ansiClrScr(void);
void ansiEraseLine(void);
void ansiCursorPos(unsigned char, unsigned char);
void ansiCursorSav(void);
void ansiCursorRst(void);
void scrSetup(void);
void gpibCalc(gcmd *);



//End file.
